﻿using System;
using System.Collections.Generic;
using System.Text;
// 
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace playwav
{
    enum Tag
    {
        ON,
        OFF,
        END,
    }


    class Note
    {
        public readonly Tag m_tag;
        public readonly double m_timing;  //  １拍が１
        public readonly double m_note;    //  440.0

        public Note(Tag tag, double timing, double note)
        {
            m_tag = tag;
            m_timing = timing;
            m_note = note;
        }

        public void Disp()
        {
            Debug.WriteLine("(" + m_tag.ToString() + "," + m_timing.ToString() + "," + m_note.ToString() + ")");
        }    
    }

    class SeqData
    {
        static double NoteA = 440.0;
        static double NoteC = NoteA * 1.189207;
        static double NoteD = NoteC * 1.122462;
        static double NoteE = NoteC * 1.259921;
        static double NoteF = NoteC * 1.334840;
        static double NoteG = NoteC * 1.498307;
        static double NoteA2 = NoteA*2.0;
        static double NoteB2 = NoteA2 * 1.122462;
        static double NoteC2 = NoteC * 2.0;
        double m_bps;   　//  120.0 bps
        double m_bps_second;　//  一拍の時間
        List<Note> m_notes;

        public void Disp()
        {
            Debug.WriteLine("m_bps=" + m_bps.ToString());
            foreach(Note n in m_notes)
            {
                n.Disp();
            }

        }

        public SeqData()
        {

        }

        static public SeqData GetSampleKyoku()
        {
            SeqData seqdata = new SeqData();

            seqdata.m_bps = 120.0;
            seqdata.m_bps_second = 60.0 / seqdata.m_bps;           //　一拍の時間（秒）
            seqdata.m_notes = new List<Note>();

            double cnt = 0.0;

#if false
            seqdata.m_notes.Add(new Note(Tag.ON, 0.0, NoteA * 16.0)); 
            seqdata.m_notes.Add(new Note(Tag.ON, 0.05, 0.0));
            seqdata.m_notes.Add(new Note(Tag.OFF, 0.1, 0.0));
            seqdata.m_notes.Add(new Note(Tag.ON,  0.15, NoteA * 4.0));
            seqdata.m_notes.Add(new Note(Tag.END, 0.2, NoteA));
#endif
#if true
            for (int i = 0; i < 8; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA * 4.0)); cnt += 0.125;
                seqdata.m_notes.Add(new Note(Tag.OFF, cnt, NoteA )); cnt += (1.0 - 0.125);
            }

            for (int i = 0; i < 32; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA * 4.0)); cnt += 0.125 * 0.2;
                seqdata.m_notes.Add(new Note(Tag.OFF, cnt, NoteA * 4.0)); cnt += (1.0 - 0.125) * 0.2;
            }
 
            // 8.0
            for (int i = 0; i < 1; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA));
                seqdata.m_notes.Add(new Note(Tag.ON, cnt + 0.1, NoteA * 4.0));
                seqdata.m_notes.Add(new Note(Tag.OFF, cnt + 2.0 - 0.1, NoteA * 4.0));
                cnt += 2.0;
            }
            for (int i = 0; i < 4; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA * 8.0)); cnt += 0.125*0.5;
                seqdata.m_notes.Add(new Note(Tag.OFF, cnt, NoteA )); cnt += (1.0 - 0.125)*0.5;
            }

            double ratio = 1.0;
            double tempo = 0.5;
            for (int i = 0; i < 8; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, ratio * NoteC)); cnt += tempo;
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, ratio * NoteC * 4.0)); cnt += tempo;
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, ratio * NoteD)); cnt += tempo;
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, ratio * NoteE * 8.0)); cnt += tempo;
                ratio *= 1.334840;
            }
            for (int i = 0; i < 8; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA * 0.5)); cnt += 0.25;
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA * 2.0)); cnt += 0.25;
            }
            for (int i = 0; i < 16; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA)); cnt += 0.125;
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteC)); cnt += 0.125;
            }
            seqdata.m_notes.Add(new Note(Tag.END, cnt, NoteC*8.0));
#endif

            return (seqdata);
        }

        static public SeqData GetSamplePipipi()
        {
            SeqData seqdata = new SeqData();

            seqdata.m_bps = 120.0;
            seqdata.m_bps_second = 60.0 / seqdata.m_bps;           //　一拍の時間（秒）
            seqdata.m_notes = new List<Note>();

            double cnt = 0.0;
            for (int i = 0;i<2; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA * 8.0)); cnt += 0.125 * 0.2;
                seqdata.m_notes.Add(new Note(Tag.OFF, cnt, NoteA * 0.50)); cnt += (1.0 - 0.125) * 0.2;
            }
            seqdata.m_notes.Add(new Note(Tag.END, cnt, NoteA*4.0)); cnt += 0.5;
            return (seqdata);
        }

        static public SeqData GetSampleDoremi()
        {
            SeqData seqdata = new SeqData();

            seqdata.m_bps = 120.0;
            seqdata.m_bps_second = 60.0 / seqdata.m_bps;           //　一拍の時間（秒）
            seqdata.m_notes = new List<Note>();
            double cnt = 0.0;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteC)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteD)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteE)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteF)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteG)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA2)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteB2)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteC2)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.END, cnt, NoteC2));

            return (seqdata);
        }

        static public SeqData GetSampleKaeru()
        {
            SeqData seqdata = new SeqData();

            seqdata.m_bps = 120.0;
            seqdata.m_bps_second = 60.0 / seqdata.m_bps;           //　一拍の時間（秒）
            seqdata.m_notes = new List<Note>();
            double cnt = 0.0;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteC)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteD)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteE)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteF)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteE)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteD)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteC)); cnt += 0.5;
            seqdata.m_notes.Add(new Note(Tag.END, cnt, NoteC));

            return (seqdata);
        }

        static public SeqData GetSampleRa3()
        {
            SeqData seqdata = new SeqData();

            seqdata.m_bps = 120.0;
            seqdata.m_bps_second = 60.0 / seqdata.m_bps;           //　一拍の時間（秒）
            seqdata.m_notes = new List<Note>();
            double cnt = 0.0;
            for (int i = 0; i < 8; i++)
            {
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA*0.5)); cnt += 0.25;
                seqdata.m_notes.Add(new Note(Tag.ON, cnt, NoteA*2.0)); cnt += 0.25;
            }
            seqdata.m_notes.Add(new Note(Tag.END, cnt, NoteA * 0.5));

            return (seqdata);
        }

        /// <summary>
        /// レンダリングする
        /// とりあえずダブルでやっとく
        /// とりあえずモノラルでやっとく
        /// </summary>
        /// <returns></returns>
        public List<double> ToNotes(double onesample)
        {
            double lasttiming = m_notes[m_notes.Count - 1].m_timing;
            // 最後の時間
            double lasts = lasttiming * m_bps_second;
            double dsamples = lasts / onesample;
            // hako youi suru
            List<double> ld = new List<double>();

            //  一旦全部ノートに直す
            int index=0;
            double nextsecond = m_notes[index].m_timing * m_bps_second;
            double note = 0.0;
            for (int i = 0; ; i++)
            {
                double nowsecond = onesample * i;
                if (nowsecond >= nextsecond)
                {
                    note = m_notes[index].m_note;
                    if (m_notes[index].m_tag == Tag.END)
                    {
                        break;
                    }
                    index++;
                    Debug.WriteLine("index=" + index.ToString());
                    nextsecond = m_notes[index].m_timing * m_bps_second;
                }
                ld.Add(note);
            }


            return (ld);
        }

        /// <summary>
        /// レンダリングする
        /// とりあえずダブルでやっとく
        /// とりあえずモノラルでやっとく
        /// </summary>
        /// <returns></returns>
        public List<double> ToNotes2(double onesample)
        {
            // 一旦小節データを秒に直す
            double[] seconds = new double[m_notes.Count];
            for (int i = 0; i < m_notes.Count; i++)
            {
                seconds[i] = m_notes[i].m_timing * m_bps_second;
            }

            List<double> ld = new List<double>();
            int ind = 0;
            for (int i = 0; ; i++)
            {
                double nowsecond = onesample * i;
                if (nowsecond > seconds[ind + 1])
                {
                    ind++;
                    if (m_notes[ind].m_tag == Tag.END)
                    {
                        break;
                    }
                }
                double maenote = m_notes[ind].m_note;
                double atonote = m_notes[ind + 1].m_note;
                double maesecond = seconds[ind];
                double atosecond = seconds[ind + 1];

                if (m_notes[ind].m_tag == Tag.ON)
                {
                    double note =
                        (atonote - maenote) * (nowsecond - maesecond) / (atosecond - maesecond) + maenote;
                    ld.Add(note);
                }
                else
                {
                    ld.Add(0.0);
                }

            }
            return (ld);
        }
    }
}
